/* istanbul ignore file: it's obviously simple. */
const noop = () => { };
const identity = (x) => x;
function doCleanup(cleanup) {
    if (typeof cleanup === 'function') {
        cleanup();
    }
    else if (typeof cleanup === 'object') {
        if ('abort' in cleanup) {
            cleanup.abort();
        }
        else if ('cancel' in cleanup) {
            cleanup.cancel();
        }
        else if ('unsubscribe' in cleanup) {
            cleanup.unsubscribe();
        }
        else if ('dispose' in cleanup) {
            cleanup.dispose();
        }
    }
}
function hostReportErrors(error) {
    setTimeout(() => {
        throw error;
    });
}

const Symbol_observable = 
/* istanbul ignore next: simple expression*/ (() => (typeof Symbol === 'function' && Symbol.observable) || '@@observable')();

class TrackContext {
    constructor(observer) {
        this.observer = observer;
        this.lastLog = null;
        this.cursor = this.observer.fs;
        this.clean = null;
    }
    track(data) {
        if (data.flags & 16 /* Evalutating */) {
            if (data === this.observer) {
                if (data.flags & 16384 /* ValueWasError */)
                    throw data.hs; // [EXIT 7]
                return data.hs; // [EXIT 6]
            }
            throw new ReferenceError('A circular reference occurred.'); // [EXIT 5]
        }
        this.logSource(data);
        // if propagating and current are still markForCheck
        if (data.flags & 32 /* MarkForCheck */) {
            // currently if MarkForCheck it must be in propagate phase
            throw DEFER_COMPUTATION; // [EXIT 4]
        }
        if (data.flags & 1024 /* Stale */) {
            estimate(data); // maybe [EXIT 3]
        }
        if (data.flags & 2048 /* ValueIsError */) {
            throw data.value; // [EXIT 2]
        }
        return data.value; // current value [EXIT 1]
    }
    logSource(data) {
        if (this.lastLog) {
            if (this.lastLog === data) {
                return;
            }
        }
        this.lastLog = data;
        if (this.cursor === null) {
            insertNewSource(this.observer, data);
        }
        else {
            if (this.cursor.source !== data) {
                this.clean = this.observer.ls;
                this.observer.ls = this.cursor.prev;
                if (this.observer.ls) {
                    this.observer.ls.next = null; // new last_source
                    this.cursor.prev = null; //
                }
                else {
                    this.observer.fs = null;
                }
                this.cursor = null;
                insertNewSource(this.observer, data);
            }
            else {
                this.cursor = this.cursor.next;
            }
        }
    }
    cleanup() {
        if (this.cursor !== null) {
            this.clean = this.observer.ls;
            this.observer.ls = this.cursor.prev;
            if (this.observer.ls) {
                this.observer.ls.next = null; // new last_source
                this.cursor.prev = null;
            }
            else {
                // cursor.prev == null => zero dependency is logged: settled forever.
                this.observer.fs = null;
            }
        }
        return this.clean;
    }
}
class EvalContext {
    constructor() {
        this.toCleanup = [];
    }
    cleanup() {
        for (const c of this.toCleanup) {
            cleanupSources(c, null);
        }
    }
    /**
     * Evaluate the expression and track (refresh) dependencies.
     * @param observer
     * @param expression
     * @returns Computed latest vaue
     * @throws User-land errors, DEFER_COMPUTATION
     */
    evaluate(observer, expression) {
        const ctx = new TrackContext(observer);
        observer.flags |= 16 /* Evalutating */;
        try {
            return expression((x) => ctx.track(x));
        }
        finally {
            observer.flags &= ~16 /* Evalutating */;
            const clean = ctx.cleanup();
            if (clean)
                this.toCleanup.push(clean);
        }
    }
}
function setData(data, value) {
    if (!ct) {
        return nt.start(() => setData(data, value));
    }
    ct.setData(data, value);
}
function accessCurrent(data, obs = null) {
    if (data.flags & 16 /* Evalutating */) {
        if (data === obs) {
            if (data.flags & 16384 /* ValueWasError */)
                throw data.hs;
            return data.hs;
        }
        throw new ReferenceError('A circular reference occurred.');
    }
    if (data.flags & 32 /* MarkForCheck */ && ct !== null && !ct.flushing) {
        ct.flush();
    }
    if (data.flags & 1024 /* Stale */) {
        // eager evaluate latest value
        try {
            data.flags |= 16 /* Evalutating */;
            return data.c((x) => accessCurrent(x, data));
        }
        finally {
            data.flags &= ~16 /* Evalutating */;
        }
    }
    else {
        if (data.flags & 2048 /* ValueIsError */) {
            throw data.value; // [EXIT 2]
        }
        return data.value; // current value [EXIT 1]
    }
}
function estimate(memo) {
    memo.flags &= ~1024 /* Stale */;
    const ctx = new EvalContext();
    try {
        memo.value = ctx.evaluate(memo, memo.c);
        memo.flags &= ~2048 /* ValueIsError */;
    }
    catch (e) {
        if (e === DEFER_COMPUTATION) {
            // it will be propagated later
            // _must mark dirty_ because the source may be unchanged but this node must be recomputed.
            memo.flags |= 32 /* MarkForCheck */ | 4 /* Dirty */ | 1024 /* Stale */;
            memo.dc++;
            throw e; // [EXIT 3]
        }
        memo.flags |= 2048 /* ValueIsError */;
        memo.value = e;
    }
    ctx.cleanup(); //should be just no-op
}
function cleanupObserver(memo) {
    if (memo.ls === null) {
        return;
    }
    cleanupSources(memo.ls, null);
    memo.fs = null;
    memo.ls = null;
}
function cleanupSources(last, until) {
    do {
        const { source, observer_ref: { next, prev }, } = last; /* no first source so do nothing */
        // last.observer_ref.observer = null; // not necessary (but might be gc friendly)
        (next ? (next.prev = prev) : (source.lo = prev))
            ? (prev.next = next)
            : null;
        if ((source.flags & 8194 /* ManagedMemo */) === 8194 /* ManagedMemo */ &&
            source.lo === null) {
            source.flags = (source.flags | 1024 /* Stale */) & ~32 /* MarkForCheck */;
            source.dc = 0; // not usable...
            source.value = null;
            // it's last observer // but propagation might tackle this.
            cleanupObserver(source);
        }
    } while ((last = last.prev) !== until);
}
function isMemo(node) {
    return (node.flags & 2 /* Memo */) === 2 /* Memo */;
}
function markObserversForCheck(stack) {
    while (stack.length > 0) {
        let lo = stack.pop().lo;
        while (lo !== null) {
            const observer = lo.observer;
            observer.dc++;
            if ((observer.flags & 32 /* MarkForCheck */) === 0) {
                observer.flags |= 32 /* MarkForCheck */;
                if (observer.flags & 2 /* Memo */) {
                    stack.push(observer);
                }
            }
            lo = lo.prev;
        }
    }
}
let ct = null;
class TransactionContext {
    constructor() {
        this.dirty = [];
        this.effects = [];
        this.flushing = false;
    }
    start(fn) {
        const stored = ct;
        ct = this;
        let retValue = null;
        try {
            retValue = fn();
            this.flush();
        }
        finally {
            ct = stored;
        }
        controlOnCommit(() => this.commit());
        return retValue;
    }
    flush() {
        this.flushing = true;
        this.propagate(this.dirty);
        this.flushing = false;
    }
    commit() {
        const effects = this.effects;
        while (effects.length) {
            const x = effects.pop();
            x.flags &= ~64 /* EffectWillCommit */;
            /**
             * should catch error here?
             * no. and unrecoverable.
             */
            x.e();
        }
    }
    setData(data, value) {
        /* istanbul ignore if: noop */ if (data.cp(data.value, value)) {
            return;
        }
        data.value = value;
        if (data.flags & 8 /* Changed */) {
            return;
        }
        data.flags |= 8 /* Changed */;
        markObserversForCheck([data]); //
        this.dirty.push(data);
    }
    propagate(stack) {
        const ctx = new EvalContext();
        while (stack.length > 0) {
            const data = stack.pop();
            const sdf = data.flags;
            const dirtyIfChanged = (sdf & 8 /* Changed */) >> 1; // 8>>1 = 4
            data.flags &= ~8 /* Changed */;
            let lo = data.lo;
            while (lo !== null) {
                const observer = lo.observer;
                lo = lo.prev;
                observer.flags |= dirtyIfChanged; // changes -> dirty
                observer.dc--;
                if (observer.dc === 0) {
                    const scf = (observer.flags &= ~32 /* MarkForCheck */);
                    if (isMemo(observer)) {
                        // it's a memo
                        stack.push(observer); // push()
                        if (scf & 4 /* Dirty */) {
                            observer.flags =
                                ((scf & 2048 /* ValueIsError */) > 0
                                    ? observer.flags | 16384 /* ValueWasError */
                                    : observer.flags & ~16384 /* ValueWasError */) & ~4 /* Dirty */;
                            observer.hs = observer.value;
                            try {
                                const currentValue = ctx.evaluate(observer, observer.c);
                                // compare value , if changed, mark as changed
                                if (!observer.cp(currentValue, observer.value) ||
                                    observer.flags & 2048 /* ValueIsError */) {
                                    observer.flags =
                                        (observer.flags | 8 /* Changed */) & ~2048 /* ValueIsError */;
                                    observer.value = currentValue;
                                }
                            }
                            catch (e) {
                                if (e === DEFER_COMPUTATION) {
                                    // computation deferred as a upstream dependency is waiting for propagation.
                                    stack.pop(); // pop()
                                    observer.flags |= 4 /* Dirty */ | 32 /* MarkForCheck */;
                                    observer.dc++;
                                    continue;
                                }
                                // TODO: should check error equality?
                                observer.value = e;
                                observer.flags |= 8 /* Changed */ | 2048 /* ValueIsError */;
                            }
                        }
                    }
                    else {
                        // if (scf & BitFlags.Effect) {
                        if (scf & 4 /* Dirty */) {
                            observer.flags &= ~4 /* Dirty */;
                            if (~scf & 64 /* EffectWillCommit */) {
                                this.effects.push(observer);
                                observer.flags |= 64 /* EffectWillCommit */;
                            }
                        }
                        // }
                    }
                }
            }
        }
        ctx.cleanup();
    }
}
const nt = new TransactionContext();
function batch(fn) {
    if (ct !== null) {
        return fn();
    }
    return nt.start(fn);
}
function insertNewSource(observer, source) {
    observer.ls = observer.ls
        ? (observer.ls.next = {
            prev: observer.ls,
            next: null,
            source,
            observer_ref: insertNewObserver(source, observer),
        })
        : (observer.fs = {
            prev: null,
            next: null,
            source,
            observer_ref: insertNewObserver(source, observer),
        });
}
function insertNewObserver(source, observer) {
    return (source.lo = source.lo
        ? (source.lo.next = {
            prev: source.lo,
            next: null,
            observer,
        })
        : { prev: null, next: null, observer });
}
function createReaction(onCommit) {
    return {
        flags: 256 /* Effect */,
        dc: 0,
        fs: null,
        ls: null,
        e: onCommit,
    };
}
function executeReaction(reaction, fn) {
    const ctx = new EvalContext();
    try {
        return ctx.evaluate(reaction, fn);
    }
    finally {
        ctx.cleanup();
    }
}
function createData(value, comp = Object.is) {
    return {
        flags: 1 /* Data */,
        lo: null,
        value: value,
        cp: comp,
    };
}
function createMemo(fn, comp = Object.is, initial) {
    return {
        flags: 2 /* Memo */ | 1024 /* Stale */ | 8192 /* Managed */,
        lo: null,
        value: null,
        cp: comp,
        dc: 0,
        fs: null,
        ls: null,
        c: fn,
        hs: initial,
    };
}
let controlOnCommit = ((x) => x());
/* istanbul ignore next: simple */
function takeControlOnCommit(controller) {
    controlOnCommit = controller;
}
/**
 * Only for _extra dependencies added_
 */
const DEFER_COMPUTATION = {};

var _a$1;
class Cell {
    /**  @internal */
    constructor(/**  @internal */ internal) {
        this.internal = internal;
        this[_a$1] = 'Cell';
    }
    static of(initial) {
        return new Cell(createData(initial));
    }
    get current() {
        return accessCurrent(this.internal);
    }
    [Symbol_observable]() {
        return this;
    }
    map(fn) {
        return new Cell(createMemo(($) => fn($(this.internal))));
    }
    /**
     * @deprecated Should be only used for interop.
     */
    subscribe(...params) {
        return new Observable((observer) => {
            const reaction = new Reaction(() => {
                try {
                    observer.next(this.current);
                }
                catch (e) {
                    observer.error(e);
                }
            });
            observer.next(this.current);
            reaction.track((t) => t(this));
            return () => reaction.dispose();
        }).subscribe(...params);
    }
}
_a$1 = Symbol.toStringTag;
function createTrackFn$1($) {
    const ret = (cell) => $(cell.internal);
    Object.defineProperties(ret, {
        error: {
            value: (cell) => {
                try {
                    ret(cell);
                    return null;
                }
                catch (e) {
                    return e;
                }
            },
        },
    });
    return ret;
}
class Reaction {
    constructor(callback) {
        this.internal = createReaction(callback);
    }
    /**
     *
     * @param program
     * @returns return value of `program`
     * @throws User-land errors in program()
     */
    track(program) {
        return executeReaction(this.internal, ($) => program(createTrackFn$1($)));
    }
    dispose() {
        cleanupObserver(this.internal);
    }
}
function mutable(initialValue, options) {
    const internal = createData(initialValue, options === null || options === void 0 ? void 0 : options.comparator);
    return [
        new Cell(internal),
        (v) => {
            if (v instanceof Function) {
                setData(internal, v(internal.value));
            }
            else {
                setData(internal, v);
            }
        },
    ];
}
function computed(expr, initial, options) {
    return new Cell(createMemo((t) => expr(createTrackFn$1(t)), options === null || options === void 0 ? void 0 : options.comparator, initial));
}
const d = (cell, track) => track(cell);
function combined(obj) {
    if (obj instanceof Array) {
        const elementMap = obj.map((item) => (item instanceof Cell ? d : identity));
        return computed(($) => elementMap.map((x, i) => x(obj[i], $)));
    } /* istanbul ignore else: simple */
    else if (typeof obj === 'object' &&
        obj !== null) {
        const entryMap = Object.entries(obj).map(([key, value]) => (value instanceof Cell ? [key, d, value] : [key, identity, value]));
        return computed(($) => Object.fromEntries(entryMap.map(([key, map, value]) => {
            return [key, map(value, $)];
        })));
    }
    // istanbul ignore next: simple
    throw new TypeError(undefined);
}

var _a;
class CellSuspended {
    constructor(executor) {
        this.executor = executor;
    }
}
const DEFER_SUSPENSION = new CellSuspended(
/* istanbul ignore next: not used */ () => new Promise(noop));
function createTrackFn($) {
    const ret = (cell) => $(cell.internal);
    Object.defineProperties(ret, {
        error: {
            value: (cell) => {
                try {
                    ret(cell);
                    return null;
                }
                catch (e) {
                    return e;
                }
            },
        },
        suspend: {
            value: (cell) => {
                const value = ret(cell);
                if (cell._currentSuspension !== null) {
                    throw DEFER_SUSPENSION;
                }
                return value;
            },
        },
    });
    return ret;
}
class SuspendedCell extends Cell {
    /** @internal */
    constructor(expr, options) {
        super(createMemo(($) => {
            try {
                $(this._notifier);
                const value = expr(createTrackFn($));
                this._currentSuspension = null;
                return value;
            }
            catch (e) {
                if (e instanceof CellSuspended) {
                    this._currentSuspension = e;
                    //
                    if ((this.internal.flags & 1024 /* Stale */) === 0) {
                        // TODO: isn't it abusing?
                        // estimate or propagate
                        this.cancelLatest();
                        this.forkCurrent(e);
                    }
                    return fallback;
                }
                throw e; // throw user-land error, or DEFER_COMPUTATION
            }
        }, options === null || options === void 0 ? void 0 : options.comparator, options === null || options === void 0 ? void 0 : options.initial));
        /** @internal */
        this._currentSuspension = null;
        /** @internal */
        this._executing = null;
        this[_a] = 'SuspendedCell';
        this._notifier = createData(0);
        const fallback = options === null || options === void 0 ? void 0 : options.fallback;
    }
    /** @internal */
    cancelLatest() {
        if (this._executing !== null) {
            // cancel current?
            this._executing.abort();
            this._executing = null;
        }
    }
    /** @internal */
    forkCurrent(s) {
        if (s === DEFER_SUSPENSION) {
            return; // control by upstream
        }
        const cc = (this._executing = new AbortController());
        s.executor(cc.signal).finally(() => {
            if (this._executing !== cc)
                return; // not needed.
            this._executing = null;
            setData(this._notifier, this._notifier.value + 1); // maybe schedule it?
        });
    }
}
_a = Symbol.toStringTag;
function suspended(expr, options) {
    return new SuspendedCell(expr, options);
}
// const queue: (() => void)[] = [];
// function enqueueJob(job: () => void) {
//   if (queue.length === 0) {
//     enqueueTask(() => {
//       batch(() => {
//         while (queue.length) {
//           queue.pop()!();
//         }
//       });
//     });
//   }
//   queue.push(job);
// }
// const enqueueTask = /* istanbul ignore next */ (
//   (() => {
//     if (typeof setImmediate === 'function') {
//       return setImmediate;
//     }
//     if (typeof MessageChannel !== 'undefined') {
//       const channel = new MessageChannel();
//       return (callback) => {
//         channel.port1.onmessage = callback;
//         channel.port2.postMessage(undefined);
//       };
//     } /* istanbul ignore next */
//     return (callback) => setTimeout(callback, 0);
//   }) as () => (callback: () => void) => void
// )();

let currentCollecting = null;
function collectScope() {
    const stored = currentCollecting;
    currentCollecting = [];
    const current = currentCollecting;
    return () => {
        if (currentCollecting !== current) {
            throw new Error('Nested scope');
        }
        currentCollecting = stored;
        return new LifecycleScope(current);
    };
}
class LifecycleScope {
    constructor(onmountLogics) {
        this.onmountLogics = onmountLogics;
        this.attached = false;
        this.detach = noop;
    }
    attach() {
        /* istanbul ignore if */
        if (this.attached) {
            return;
        }
        this.attached = true;
        const cleanups = this.onmountLogics.map((x) => {
            const cleanup = x();
            return cleanup;
        });
        return (this.detach = () => {
            if (this.attached) {
                this.attached = false;
                cleanups.forEach(doCleanup);
            }
        });
    }
}
function lifecycle(onInit) {
    if (currentCollecting) {
        currentCollecting.push(onInit);
        return;
    }
    throw new TypeError('Not inside a scope');
}

const Identifier = {
    of: (description, options) => {
        const id = Symbol(description);
        if (options) {
            if (options.suggesstion) ;
        }
        return id;
    },
};
let currentContext = null;
class Context {
    constructor(entries = new Map()) {
        this.entries = entries;
    }
    get(token, options) {
        var _a;
        if (token === CONTEXT) {
            // FIXME: wrong type
            return this;
        }
        if (this.entries.has(token)) {
            return this.entries.get(token);
        }
        if (options === null || options === void 0 ? void 0 : options.optional) {
            return (_a = options.defaultValue) !== null && _a !== void 0 ? _a : undefined;
        }
        throw new IdentifierNotFoundError(token);
    }
    runInContext() {
        const stored = currentContext;
        currentContext = this;
        return () => {
            if (currentContext !== this) {
                throw new Error('Exit inner context before outer context.');
            }
            currentContext = stored;
        };
    }
    inherit(module) {
        return new Context(new Map([
            ...this.entries,
            ...Object.getOwnPropertySymbols(module).map((ssymbol) => [ssymbol, module[ssymbol]]),
        ]));
    }
    /**
     * Build a new context based on current from concern.
     * @param concern
     * @returns
     */
    build(concern) {
        var _a;
        const exitContext = this.runInContext();
        try {
            const exports = (_a = concern()) !== null && _a !== void 0 ? _a : {};
            return this.inherit(exports);
        }
        finally {
            exitContext();
        }
    }
}
Context.EMPTY = new Context();
class IdentifierNotFoundError extends Error {
    constructor(identifier) {
        super(`'${identifier.description}' is not found in current context.`);
        this.identifier = identifier;
    }
}
function injected(token, options) {
    if (currentContext) {
        return currentContext.get(token, options);
    }
    throw new TypeError('Not inside a context.');
}
function reduceConcerns(concerns) {
    return () => {
        var _a;
        const exports = {};
        let context = currentContext; // TODO: maybe not?
        for (const concernOrArray of concerns) {
            const concern = concernOrArray instanceof Array
                ? reduceConcerns(concernOrArray)
                : concernOrArray;
            const exitContext = context.runInContext();
            try {
                const concernExport = (_a = concern()) !== null && _a !== void 0 ? _a : {};
                Object.assign(exports, concernExport);
                context = context.inherit(concernExport);
            }
            finally {
                exitContext();
            }
        }
        return exports;
    };
}
const CONTEXT = Identifier.of('Context');
const CONCERN_HOC_FACTORY = Identifier.of('ConcernHighOrderComponentFactory');

/* istanbul ignore file: simple */
class Reference {
    constructor(_current) {
        this._current = _current;
    }
    get current() {
        return this._current;
    }
    static create(initial) {
        const ref = new Reference(initial);
        const binder = (value) => {
            ref._current = value;
        };
        Object.defineProperties(binder, {
            bind: {
                get: () => binder,
                set: (value) => binder(value),
            },
        });
        binder[Symbol_bind_reference] = true;
        return [ref, binder];
    }
}
const Symbol_bind_reference = Symbol('reference_setter');
function isReferenceSetter(value) {
    return (typeof value === 'function') && value[Symbol_bind_reference] === true;
}
function reference(initial) {
    return Reference.create(initial);
}

function effect(sideEffect) {
    const doEffect = ($) => {
        try {
            sideEffect($);
        }
        catch (e) {
            console.error(`Captured error inside effect().`, e);
        }
    };
    const reaction = new Reaction(() => {
        reaction.track(doEffect);
    });
    lifecycle(() => {
        reaction.track(doEffect);
        return () => reaction.dispose();
    });
}
// export class LayoutQueue {
//   private queue: (() => void)[] = [];
//   enqueue(job: () => void) {
//     if (this.queue.length === 0) {
//       requestAnimationFrame(() => {
//         if (!this.willFlush /* template will take care */ && this.queue.length) {
//           this.flush();
//         }
//       });
//     }
//     this.queue.push(job);
//   }
//   private willFlush: boolean = false;
//   preflush() {
//     this.willFlush = true;
//   }
//   flush() {
//     this.willFlush = false;
//     try {
//       for (const job of this.queue) {
//         job();
//       }
//     } finally {
//       this.queue.length = 0;
//     }
//   }
// }
// export const LAYOUT_QUEUE = Identifier.of<LayoutQueue>(LayoutQueue.name);
// export function layout(sideEffect: () => void) {
//   const queue = consume(LAYOUT_QUEUE);
//   const doEffect = () => {
//     try {
//       sideEffect();
//     } catch (e) {
//       console.error(`Captured error inside layout().`, e);
//     }
//   };
//   const reaction = new Reaction(() => {
//     queue.enqueue(() => reaction.execute(doEffect));
//   });
//   lifecycle(() => {
//     reaction.execute(doEffect);
//     return () => reaction.dispose();
//   });
// }

function isCell(value) {
    return value instanceof Cell;
}

export { CONCERN_HOC_FACTORY, CONTEXT, Cell, CellSuspended, Context, Identifier, IdentifierNotFoundError, LifecycleScope, Reaction, Reference, SuspendedCell, Symbol_observable, takeControlOnCommit as UNSTABLE_takeControlOnCommit, batch, collectScope, combined, computed, doCleanup, effect, hostReportErrors, identity, injected, isCell, isReferenceSetter, lifecycle, mutable as mut, mutable, noop, reduceConcerns, reference as ref, reference, suspended };
